#include <iostream>
#include <array>
#include <random>
#include <cmath> 
#include <SDL2/SDL.h>
#include <utility>
#include <cstdlib>


#define RESY 100
#define RESX 100
#define MAXITER (RESX*RESY*iter_multiplier)
using namespace std;

static double T = 245.0;

#define phys_beta (1.0/T)

static double J = 1.0;
static double h = 0.0;

enum Spin {
    up = 1,
    down = -1,
};

inline void flip(Spin &spin)
{
    spin = (spin == up) ? down : up;
}
// we add a line of zeroes around the area to simplify the computation kernel
// this way, we won't need a series of ifs in the loop or breaking up the loop
typedef array<array<Spin, RESX>, RESY> Field;

double get_energy_difference_for_flip(const Field& field, unsigned i, unsigned j)
{
    int sum = 0;
    // one up
    sum += (i != 0)        ? field[i - 1][j] : 0;
    // one down
    sum += (i != RESX - 1) ? field[i + 1][j] : 0;
    // one left
    sum += (j != 0)        ? field[i][j - 1] : 0;
    // one right
    sum += (j != RESY - 1) ? field[i][j + 1] : 0;
    
    return 2 * J * field[i][j] * sum + 2 * h * field[i][j];
}

pair<double, double> get_results(const Field& field)
{
    int mag = 0;
    int pairwise = 0;
    
    // everthing except the last column and row
    for(int y = 0; y < (RESY - 1); y++)
    {
	for(int x = 0; x < (RESX - 1); x++)
	{
	    pairwise += field[y][x]*field[y + 1][x];
	    pairwise += field[y][x]*field[y][x + 1];
	    mag += field[y][x];
	}
    }
    
    // last column
    for(int y = 0; y < (RESY - 1); y++)
    {
	pairwise += field[y][RESX - 1]*field[y + 1][RESX - 1];
	mag += field[y][RESX - 1];
    }
    
    // last row
    for(int x = 0; x < (RESX - 1); x++)
    {
	pairwise += field[RESY - 1][x]*field[RESY - 1][x + 1];
	mag += field[RESY - 1][x];

    }
    
    // last field
    mag += field[RESY - 1][RESX - 1];

    double magnetization = static_cast<double>(mag) / RESX / RESY;

    return pair(- J * pairwise + h * mag, magnetization);
}

int main(int argc, char *argv[] ) {
    
    T = atof(argv[1]);
    int iter_multiplier = atoi(argv[2]);
    //h = atof(argv[1]);
    // stolened right from the reference
    random_device rd;
    mt19937 gen(rd());
    uniform_int_distribution<> idis(0, RESX * RESY - 1);
    uniform_real_distribution<> fdis(0.0, 1.0);
    uniform_int_distribution<> spindis(0, 1);
    
    Field field;
    for(int y = 0; y < RESY; y++)
    {
	for(int x = 0; x < RESX; x++)
	{
	    field[y][x] = ((spindis(gen) == 1) ? up : down);
	}
    }
    
    double dE;
    
    int xy, x, y;
    double threshold;
    for(int i = 0; i < MAXITER; i++)
    {
	// maybe i could use 2 random numbers but I can't be arsed to created 2 uniform generators
	xy  = idis(gen);
	x = xy % RESX;
	y = xy / RESX;
	dE = get_energy_difference_for_flip(field, y, x);
	if(dE < 0)
	{
	    flip(field[y][x]);
	    continue;
	}
	threshold = fdis(gen);
	if(exp(- phys_beta * dE) < threshold)
	{
	    flip(field[y][x]);
	    continue;
	}
    }

    // SDL_Window *window;
    // SDL_Renderer *renderer;
    // SDL_Event event;

    // SDL_Init(SDL_INIT_VIDEO);
    // SDL_CreateWindowAndRenderer(RESY, RESX, 0, &window, &renderer);

    // SDL_SetRenderDrawColor(renderer, 255, 255, 255, 0);
    // SDL_RenderClear(renderer);
  
    // for(unsigned int y = 0; y < RESY; y++){
    // 	for(unsigned int x = 0; x < RESX; x++){
    // 	    SDL_SetRenderDrawColor(renderer,
    // 				   (field[y][x] == up) ? 255 : 0,
    // 				   (field[y][x] == up) ? 0 : 255,
    // 				   0,
    // 				   0);
    // 	    SDL_RenderDrawPoint(renderer, y, x); //Renders on middle of screen.
    // 	}
    // }
  
    // SDL_RenderPresent(renderer);
    // while (1) {
    // 	if (SDL_PollEvent(&event) && event.type == SDL_QUIT)
    // 	    break;
    // }
    // SDL_DestroyRenderer(renderer);
    // SDL_DestroyWindow(window);
    // SDL_Quit();
    
    pair<double, double> result = get_results(field);
    cout << T << "\t" << get<0>(result) << "\t" << get<1>(result) << endl;

    
    return EXIT_SUCCESS;
}

