#include <cmath>
#include <cstdlib>
#include <iostream>

#include <popt.h>

using namespace std;

#include "pendulum.hpp"
#include "odeint.hpp"      // ODE integration routines, Runge-Kutta ...

using namespace cpl;

const double pi = 4 * atan(1.0);

const double g = 9.8;        // acceleration of gravity

double L = 1.0;              // length of pendulum
double q = 0.5;              // damping coefficient
double Omega_D = 2.0/3.0;    // frequency of driving force
double F_D = 0.9;            // amplitude of driving force
bool nonlinear;              // linear if false

State f(const State& s) {  // extended derivative vector
    double t = s.t;
    double theta = s.theta;
    double omega = s.omega;
    State f;             // Vector with 3 components
    f.t = 1;

    f.theta = omega;
    
    f.omega = (- (g/L) * (nonlinear ? sin(theta) : theta ) -
							 q * omega + F_D * sin(Omega_D * t));
    
    return f;
}

int main() {
	const struct poptOption options_table[] = {
		{"length",   'l', POPT_ARG_DOUBLE,       &L, 0,  "Length of pendulum", NULL},
		{"damping",  'q', POPT_ARG_DOUBLE,       &q, 0, "Damping coefficient", NULL},
    {"omegaD",   'o', POPT_ARG_DOUBLE, &Omega_D, 0,    "Value of Omega_D", NULL},
		{"F_D",      'f', POPT_ARG_DOUBLE,     &F_D, 0,        "Value of F_D", NULL},
    {"theta_0",  't', POPT_ARG_DOUBLE,   &theta, 0,   "Value of theta(0)", NULL},
    {"omega_0",  '0', POPT_ARG_DOUBLE,   &omega, 0,   "Value of omega(0)", NULL},
    {"tmax",     'm', POPT_ARG_DOUBLE,    &tMax, 0,    "Integration time", NULL},
    POPT_AUTOHELP
    {NULL, 0, 0, NULL, 0, NULL, NULL}};

  poptContext opt_con = poptGetContext("sho", argc, argv, options_table, 0);
  poptSetOtherOptionHelp(opt_con, "[OPTIONS]*");
	poptGetNextOpt(opt_con);

  double dt = 0.05;
  double accuracy = 1e-6;

  double t = 0;
  State x;
  x.t = t;
  x.theta = theta;
  x.omega = omega;

  while (t < tMax) {
    adaptiveRK4Step(x, dt, accuracy, f);
    t = x.t, theta = x.theta, omega = x;
    if (nonlinear) {
      while (theta >= pi) theta -= 2 * pi;
      while (theta < -pi) theta += 2 * pi;
    }
    cout << t << '\t' << theta << '\t' << omega << '\n';
  }

  cerr << "Finished." << endl;
}

